/*  RAxML-HPC, a program for sequential and parallel estimation of phylogenetic trees 
 *  Copyright March 2006 by Alexandros Stamatakis
 *
 *  Partially derived from
 *  fastDNAml, a program for estimation of phylogenetic trees from sequences by Gary J. Olsen
 *  
 *  and 
 *
 *  Programs of the PHYLIP package by Joe Felsenstein.
 *
 *  This program is free software; you may redistribute it and/or modify its
 *  under the terms of the GNU General Public License as published by the Free
 *  Software Foundation; either version 2 of the License, or (at your option)
 *  any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 * 
 *
 *  For any other enquiries send an Email to Alexandros Stamatakis
 *  stamatak@ics.forth.gr
 *
 *  When publishing work that is based on the results from RAxML-HPC please cite:
 *  
 *  Alexandros Stamatakis: "An Efficient Program for phylogenetic Inference Using Simulated Annealing". 
 *  Proceedings of IPDPS2005,  Denver, Colorado, April 2005.
 *  
 *  AND
 *
 *  Alexandros Stamatakis, Thomas Ludwig, and Harald Meier: "RAxML-III: A Fast Program for Maximum Likelihood-based Inference of Large Phylogenetic Trees". 
 *  In Bioinformatics 21(4):456-463.
 */

#include <math.h>
#include <sys/mman.h>
#include <unistd.h>
#include <time.h> 
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "axml.h"


extern int optimizeTTinvocations;
extern int optimizeRatesInvocations;
extern int optimizeRateCategoryInvocations;
extern int optimizeAlphaInvocations;
extern int checkPointCounter;
extern int Thorough;

extern int treeStringLength;

extern double masterTime;

extern FILE   *INFILE, *permutationFile, *logFile, *infoFile;

extern char seq_file[2048];
extern char permFileName[1024], resultFileName[1024], 
  logFileName[1024], checkpointFileName[1024], infoFileName[1024], run_id[40], workdir[2048]; 

extern bestlist *bt;

void printMultipleTree(tree *tr, int quoted, char *fileName)
{
  char *str = (char *)malloc(treeStringLength);
  char *outString;
  FILE *outf = fopen(fileName, "w");
  nodeptr startp;
  startp = tr->start;

  tr->start = tr->nodep[1];
  
  treeString(str, tr, tr->start->back, 1);  
  tr->start = startp; 
  outString = str;
  while(*outString++ != '(');
  fprintf(outf, "("); 
  if(!quoted)
  {
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(outf, "%c", *outString); 
	outString++;
      }
  }
      
  free(str);
  fclose(outf);  
}



void printBootstrap(tree *tr, int quoted, char *fileName)
{
  char *str = (char *)malloc(treeStringLength);
  char *outString;
  FILE *outf = fopen(fileName, "a");
  nodeptr startp;
  startp = tr->start;

  tr->start = tr->nodep[1];
  
  treeString(str, tr, tr->start->back, 1);  
  tr->start = startp; 
  outString = str;
  while(*outString++ != '(');
  fprintf(outf, "("); 
  if(!quoted)
  {
    while(*outString != '\0')
      {
	if(*outString != '\'')
	  fprintf(outf, "%c", *outString); 
	outString++;
      }
  }
      
  free(str);
  fclose(outf);  
}






void printLogMultiple(tree *tr, analdef *adef, char *checkName, char *logName)
{
   
  if(adef->checkpoints)
    {
      char thisCheckPointName[2048], str[16];
      FILE *cpf, *lf;            
      
      lf = fopen(logName, "a");
      fprintf(lf,"%f %f %d\n", gettime() - masterTime, tr->likelihood, checkPointCounter); 
      fclose(lf);    
      
      strcpy(thisCheckPointName, checkName);
      sprintf(str, "%d", checkPointCounter);
      strcat(thisCheckPointName, str);      
      cpf = fopen(thisCheckPointName, "w");
      printTree(cpf, tr, 0);     
      fclose(cpf);    
      checkPointCounter++;
    }
  else
    {           
      FILE *lf;  
      lf = fopen(logName, "a");
      fprintf(lf,"%f %f\n", gettime() - masterTime, tr->likelihood); 
      fclose(lf);
    }    
}
    
    
  



void computeBIGRAPIDMULT (tree *tr, analdef *adef, int printLogFile, char *checkName, char *logName) 
{ 
  int i, rearrangementsMax, rearrangementsMin, noImproveCount, impr, goodTrav, bestTrav, which;                
  double lh, previousLh, difference, epsilon, currentLH, bestLH;              
  bestlist *bestT;
  
  checkPointCounter = 0;

  bestT = (bestlist *) malloc(sizeof(bestlist));
  bestT->ninit = 0;
  initBestTree(bestT, 1, tr->mxtips, tr->cdta->endsite);
    
  initInfoList(50);
 
  difference = 10.0;
  epsilon = 0.01;    
    
  Thorough = 0;           

  optimizeTTinvocations = 1;
  optimizeRatesInvocations = 1;
  optimizeRateCategoryInvocations = 1;
  optimizeAlphaInvocations = 1;
   
  
  optimizeModel(tr, adef, 1);   
  treeEvaluate(tr, 2);
  if(printLogFile)
    printLogMultiple(tr, adef, checkName, logName);

  saveBestTree(bestT, tr);   	  		      
  
  if(!adef->initialSet)
    bestTrav = determineRearrangementSetting(tr, adef, bestT);       
  else
    bestTrav = adef->initial;

  /*printf("bestTrav: %d\n", bestTrav);*/

  saveBestTree(bestT, tr);  
  impr = 1;

 
  while(impr)
    {     
      recallBestTree(bestT, 1, tr);	       
      optimizeModel(tr, adef, 1);    	      
      treeEvaluate(tr, 2);	 	  
     
      if(printLogFile)
	printLogMultiple(tr, adef, checkName, logName);
      saveBestTree(bestT, tr);	 
   
      lh = previousLh = tr->likelihood;
      
      treeOptimizeRapid(tr, 1, bestTrav, adef);
      
      impr = 0;    
      
      for(i = 1; i <= bt->nvalid; i++)
	{	 		  	   
	  recallBestTree(bt, i, tr);	    
	  treeEvaluate(tr, 0.25);	 		  
	  difference = ((tr->likelihood > previousLh)? 
			tr->likelihood - previousLh: 
			previousLh - tr->likelihood); 	    
	  if(tr->likelihood > lh && difference > epsilon)
	    {
	      impr = 1;	       
	      lh = tr->likelihood;	       
	      saveBestTree(bestT, tr);		
	    }	   	   
	}     
    }
  
  Thorough = 1;
  impr = 1;

  /*printf("SWITCHING\n");*/
  
  while(1)
    {		
      recallBestTree(bestT, 1, tr);	
     
      if(impr)
	{	    	 
	  rearrangementsMin = 1;
	  rearrangementsMax = adef->stepwidth;	    
	}			  			
      else
	{		       	   
	  rearrangementsMax += adef->stepwidth;
	  rearrangementsMin += adef->stepwidth; 	        	      
	  if(rearrangementsMax > adef->max_rearrange)	     	     	 
	    goto cleanup; 	   
	}
      
      optimizeModel(tr, adef, 1);     
      treeEvaluate(tr, 2);	    
      if(printLogFile)
	printLogMultiple(tr, adef, checkName, logName);
      previousLh = lh = tr->likelihood;	
      saveBestTree(bestT, tr);	
      
      treeOptimizeRapid(tr, rearrangementsMin, rearrangementsMax, adef);
	
      impr = 0;			      	
		
      for(i = 1; i <= bt->nvalid; i++)
	{	    		  	   

	  recallBestTree(bt, i, tr);	    
	  treeEvaluate(tr, 0.25);

	  difference = ((tr->likelihood > previousLh)? 
			tr->likelihood - previousLh: 
			previousLh - tr->likelihood); 	    
	  if(tr->likelihood > lh && difference > epsilon)
	    {
	      impr = 1;	       
	      lh = tr->likelihood;	       
	      saveBestTree(bestT, tr);		
	    }	   	   
	}	
    }
    
 cleanup:   
  freeBestTree(bestT);
  free(bestT);
  freeInfoList();
}








boolean regetdata (analdef *adef, rawdata *rdta, tree *tr)
{
  int   i, j, k, l, basesread, basesnew, ch, my_i;
  int   meaning[256];          /*  meaning of input characters */ 
  char *nameptr;
  boolean  allread, firstpass;
  char buffer[300]; 
  int len;
  getnums (rdta);

  for (i = 0; i <= 255; i++) 
    meaning[i] = -1;

  if(adef->model == M_PROTEINS || adef->model == M_PROTEINSCAT)
    {          
      meaning['A'] =  0;  /* alanine */
      meaning['R'] =  1;  /* arginine */
      meaning['N'] =  2;  /*  asparagine*/
      meaning['D'] =  3;  /* aspartic */
      meaning['C'] =  4;  /* cysteine */
      meaning['Q'] =  5;  /* glutamine */
      meaning['E'] =  6;  /* glutamic */
      meaning['G'] =  7;  /* glycine */
      meaning['H'] =  8;  /* histidine */
      meaning['I'] =  9;  /* isoleucine */
      meaning['L'] =  10; /* leucine */
      meaning['K'] =  11; /* lysine */
      meaning['M'] =  12; /* methionine */
      meaning['F'] =  13; /* phenylalanine */
      meaning['P'] =  14; /* proline */
      meaning['S'] =  15; /* serine */
      meaning['T'] =  16; /* threonine */
      meaning['W'] =  17; /* tryptophan */
      meaning['Y'] =  18; /* tyrosine */
      meaning['V'] =  19; /* valine */
      meaning['B'] =  20; /* asparagine, aspartic 2 and 3*/
      meaning['Z'] =  21; /* glutamine glutamic 5 and 6*/
      meaning['X'] =  meaning['?'] = meaning['*'] = meaning['-'] = 22; /* all = 1.0 */
    }
  else
    {      
      meaning['A'] =  1;
      meaning['B'] = 14;
      meaning['C'] =  2;
      meaning['D'] = 13;
      meaning['G'] =  4;
      meaning['H'] = 11;
      meaning['K'] = 12;
      meaning['M'] =  3;
      meaning['N'] = 15;
      meaning['O'] = 15;
      meaning['R'] =  5;
      meaning['S'] =  6;
      meaning['T'] =  8;
      meaning['U'] =  8;
      meaning['V'] =  7;
      meaning['W'] =  9;
      meaning['X'] = 15;
      meaning['Y'] = 10;
      meaning['?'] = 15;
      meaning['-'] = 15;	
    }
  
  basesread = basesnew = 0;

  allread = FALSE;
  firstpass = TRUE;
  ch = ' ';

  while (! allread) 
    {
      for (i = 1; i <= tr->mxtips; i++) 
	{   	  
	  if (firstpass) 
	    {                      	       
	      ch = getc(INFILE);
	      while(ch == ' ' || ch == '\n' || ch == '\t')
		{
		  ch = getc(INFILE);
		}

	      my_i = 0;

	      do 
		{
		  buffer[my_i] = ch;
		  ch = getc(INFILE);		   
		  my_i++;
		  if(my_i >= nmlngth)
		    {
		      printf("ERROR Names are to long, adapt constant nmlngth in\n");
		      printf("axml.h AND dnapars.c, current setting %d\n", nmlngth);
		      exit(-1);
		    }
		}
	      while(ch !=  ' ' && ch != '\n' && ch != '\t');			     	
	    }


	  j = basesread;
	  while ((j < rdta->sites) && ((ch = getc(INFILE)) != EOF) && ((! adef->interleaved) || (ch != '\n'))) 
	    {
	      uppercase(& ch);
	      if ((meaning[ch] != -1) || ch == '.') 
		{
		  j++;
		  if (ch == '.') 
		    {
		      if (i != 1) ch = rdta->y[1][j];
		      else 
			{
			  printf("ERROR: Dot (.) found at site %d of sequence 1\n", j);
			  return  FALSE;
			}
		    }
		  rdta->y[i][j] = ch;
		}
	      else 
		if(whitechar(ch) || digitchar(ch)) ;
		else 
		  {
		    printf("ERROR: Bad base (%c) at site %d of sequence %d\n",
			   ch, j, i);
		    return  FALSE;
		  }
	    }

	    if (ch == EOF) 
	      {
		printf("ERROR: End-of-file at site %d of sequence %d\n", j, i);
		return  FALSE;
	      }

	    if (! firstpass && (j == basesread)) i--; 
	    else 
	      if (i == 1) basesnew = j;
	      else 
		if (j != basesnew) 
		  {
		    printf("ERROR: Sequences out of alignment\n");		    
		    printf("%d (instead of %d) residues read in sequence %d %s\n",
			   j - basesread, basesnew - basesread, i, tr->nameList[i]);
		    return  FALSE;
		  }
	    while (ch != '\n' && ch != EOF) ch = getc(INFILE);  /* flush line */
	}                                                  /* next sequence */
      firstpass = FALSE;
      basesread = basesnew;
      allread = (basesread >= rdta->sites);
    }

 
    
  for (j = 1; j <= tr->mxtips; j++)
    {
      for (i = 1; i <= rdta->sites; i++) 
	{
	  rdta->y[j][i] = meaning[rdta->y[j][i]];		
	}
    }   	

    return  TRUE;
  } /* getdata */


void  reMakeValuesAndWeigths(tree *tr, analdef *adef)
{
  int i, j, k;
  char buf;
  double  temp, wtemp;

  if(!adef->useWeightFile)
    {
      for (i = 1; i <= tr->rdta->sites; i++) 
	tr->rdta->wgt[i] = 1;
     
      tr->cdta->wgtsum = tr->rdta->sites;   
    }
  else
    {	
      inputweights(adef, tr->rdta, tr->cdta);
    }

  INFILE = fopen(seq_file, "r");
  regetdata(adef, tr->rdta, tr);
  fclose(INFILE); 
  
  makeweights(adef, tr->rdta, tr->cdta);
  makevalues(tr->rdta, tr->cdta);
  for (j = 0; j < tr->cdta->endsite; j++) 
    {
      tr->cdta->patrat[j] = temp = 1.0;
      tr->cdta->patratStored[j] = 1.0;
      tr->cdta->rateCategory[j] = 0;
      tr->cdta->wr[j]  = wtemp = temp * tr->cdta->aliaswgt[j];
      tr->cdta->wr2[j] = temp * wtemp;
    }
}





void doBootstrap(tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[])
{
  int i, k, n;
  double  temp, wtemp, loopTime;

  n = adef->multipleRuns;
     
  strcpy(resultFileName, workdir); 
  strcat(resultFileName, "RAxML_bootstrap.");   
  strcat(resultFileName, run_id);  

  strcpy(infoFileName, workdir);
  strcat(infoFileName, "RAxML_info.");    
  strcat(infoFileName, run_id);
             
  if(filexists(infoFileName))
    {
      printf("RAxML output files with the run ID <%s> already exist \n", run_id);
      printf("in directory %s ...... exiting\n", workdir);
      exit(-1);
    }    

  makeweights(adef, rdta, cdta);
  makevalues(rdta, cdta); 

  for(i = 0; i < n; i++)
    {                      
      loopTime = gettime();

      linkdata2tree(rdta, cdta, tr);    
      initModel(tr, rdta, cdta, adef);
     
      if(i == 0)
	printModelAndProgramInfo(tr, adef, argc, argv);

      if(i == 0)
	{
	  treeStringLength = (tr->mxtips * (nmlngth+32)) + 256 + (tr->mxtips * 2);                         
	  bt = (bestlist *) malloc(sizeof(bestlist));      
	  bt->ninit = 0;
	  initBestTree(bt, 20, tr->mxtips, tr->cdta->endsite);      
	}
     
      optimizeRatesInvocations = 1;  
      optimizeTTinvocations = 1;  
      optimizeRateCategoryInvocations = 1;  
      optimizeAlphaInvocations = 1;           

      initGlobalTip(tr, adef);              

      tr->likelihood = unlikely;

      makeParsimonyTree(tr, adef);                       
      
      allocNodex(tr, adef); 	
      initrav(tr, tr->start);
      initrav(tr, tr->start->back);       
      treeEvaluate(tr, 1);          
    
      computeBIGRAPIDMULT(tr, adef, 0, checkpointFileName, logFileName);     
    
      printBootstrap(tr, 0, resultFileName);     
     
      freeNodex(tr);            
            
      
      resetBestTree(bt);
           
      for(k = 1; k <= rdta->sites; k++)
	{	  
	  rdta->wgt[k] = rdta->wgt2[k] = 1;	 	  
	  cdta->aliaswgt[k] = 0;
	}
           	      
      reMakeValuesAndWeigths(tr, adef);           
     
      loopTime = gettime() - loopTime;
      
      printf("Bootstrap[%d]: Time %f bootstrap likelihood %f\n", i, loopTime, tr->likelihood);

      infoFile = fopen(infoFileName, "a");

      fprintf(infoFile, "Bootstrap[%d]: Time %f bootstrap likelihood %f\n", i, loopTime, tr->likelihood);
     
      fclose(infoFile);     
    }
  
  printf("All bootstrapped trees written to:      %s\n"    , resultFileName);   	          
  printf("Execution information file written to:  %s\n", infoFileName);    
}


void doMultiple(tree *tr, analdef *adef, rawdata *rdta, cruncheddata *cdta, int argc, char *argv[])
{
  int i, j, n, bestI;
  double  temp, wtemp;
  double loopTime, avgTime, bestLH, avgLH;
  double *times, *likelihoods;
  char multID[1024];
  char buf[128]; 
  char temporaryFileName[2048];

  n = adef->multipleRuns;

  times = (double *)malloc(sizeof(double) * n);
  likelihoods = (double *)malloc(sizeof(double) * n);

 

  strcpy(permFileName, workdir);    
  strcpy(resultFileName, workdir); 
  strcpy(infoFileName, workdir);    
  strcat(permFileName, "RAxML_parsimonyTree.");
  strcat(resultFileName, "RAxML_result.");
 
 
  strcat(infoFileName, "RAxML_info.");    
  strcat(infoFileName, run_id);
  strcat(permFileName, run_id);
  strcat(resultFileName, run_id);
 
    
  
        
  if(filexists(infoFileName))
    {
      printf("RAxML output files with the run ID <%s> already exist \n", run_id);
      printf("in directory %s ...... exiting\n", workdir);
      exit(-1);
    }
    
  /*logFile = fopen(logFileName, "w");      */

  makeweights(adef, rdta, cdta);
  makevalues(rdta, cdta);
  linkdata2tree(rdta, cdta, tr);    

  for(i = 0; i < n; i++)
    {          
      strcpy(multID, ".RUN.");     
      sprintf(buf, "%d", i);
      strcat(multID, buf);      

      strcpy(checkpointFileName, workdir);
      strcat(checkpointFileName, "RAxML_checkpoint.");
      strcat(checkpointFileName, run_id);
      strcat(checkpointFileName, multID);
      strcat(checkpointFileName, ".");
      
      strcpy(logFileName, workdir);  
      strcat(logFileName, "RAxML_log.");
      strcat(logFileName, run_id);
      strcat(logFileName, multID);
         
      loopTime = gettime();
               
      initModel(tr, rdta, cdta, adef);       
          
      if(i == 0)
	{
	  printModelAndProgramInfo(tr, adef, argc, argv);
	  treeStringLength = (tr->mxtips * (nmlngth+32)) + 256 + (tr->mxtips * 2);                         
	  bt = (bestlist *) malloc(sizeof(bestlist));      
	  bt->ninit = 0;
	  initBestTree(bt, 20, tr->mxtips, tr->cdta->endsite);      
	}
     
      optimizeRatesInvocations = 1;  
      optimizeTTinvocations = 1;  
      optimizeRateCategoryInvocations = 1;  
      optimizeAlphaInvocations = 1;           

      initGlobalTip(tr, adef);              

      makeParsimonyTree(tr, adef);        
    
      strcpy(temporaryFileName,  permFileName);
      strcat(temporaryFileName,  multID);   
      printMultipleTree(tr, 0, temporaryFileName);
      allocNodex(tr, adef); 	    
    

      tr->likelihood = unlikely;

      initrav(tr, tr->start);
      initrav(tr, tr->start->back);          
      treeEvaluate(tr, 1);              
      computeBIGRAPIDMULT(tr, adef, 1, checkpointFileName, logFileName);           
      
      resetBestTree(bt);

      if(adef->model == M_GTRGAMMA && adef->mode != TREE_EVALUATION)
      { 

	modOpt(tr, adef);
	
	loopTime = gettime() - loopTime;
	times[i] = loopTime;
	likelihoods[i] = tr->likelihood;
	printf("RUN[%d]: Time %f, GTR+GAMMA likelihood: %f alpha shape parameter: %f\n", i, loopTime, tr->likelihood, tr->alpha);
	infoFile = fopen(infoFileName, "a");
	fprintf(infoFile, "RUN[%d]: Time %f, GTR+GAMMA likelihood: %f alpha shape parameter: %f\n", i, loopTime, tr->likelihood, tr->alpha);
	fclose(infoFile);	
	strcpy(temporaryFileName,  resultFileName);
	strcat(temporaryFileName,  multID);      
	printMultipleTree(tr, 0, temporaryFileName);
      }   

      freeNodex(tr);            
                       
      if(adef->model == M_GTRCAT || adef->model == M_GTRMIX) 
	{	 	 
	  tr->likelihood = unlikely;	
	  adef->model = M_GTRGAMMA;
	  allocNodex(tr, adef); 
	  initModel2(tr, rdta, cdta, adef);
	  optimizeRatesInvocations = 1;  
	  optimizeTTinvocations = 1;  
	  optimizeRateCategoryInvocations = 1;  
	  optimizeAlphaInvocations = 1;
	  initGlobalTip(tr, adef);
	
	  for(j = 0; j < tr->cdta->endsite; j++)
	    tr->cdta->wr[j] = tr->cdta->aliaswgt[j];
	  
	  modOpt(tr, adef);
	  
	  loopTime = gettime() - loopTime;
	  times[i] = loopTime;
	  likelihoods[i] = tr->likelihood;
	  printf("RUN[%d]: Time %f, GTR+GAMMA likelihood: %f alpha shape parameter: %f\n", i, loopTime, tr->likelihood, tr->alpha);
	  infoFile = fopen(infoFileName, "a");
	  fprintf(infoFile, "RUN[%d]: Time %f, GTR+GAMMA likelihood: %f alpha shape parameter: %f\n", i, loopTime, tr->likelihood, tr->alpha);
	  fclose(infoFile);	
	  strcpy(temporaryFileName,  resultFileName);
	  strcat(temporaryFileName,  multID);	 
	  printMultipleTree(tr, 0, temporaryFileName);	 
	  freeNodex(tr);   
	  adef->model = M_GTRCAT;	 	
	}
                      
      for (j = 0; j < tr->cdta->endsite; j++) 
	{
	  tr->cdta->patrat[j] = temp = 1.0;
	  tr->cdta->patratStored[j] = 1.0;
	  tr->cdta->rateCategory[j] = 0;
	  tr->cdta->wr[j]  = wtemp = temp * tr->cdta->aliaswgt[j];
	  tr->cdta->wr2[j] = temp * wtemp;
	}          
    }


  avgTime = 0;
  avgLH = 0;
  bestLH = unlikely;
  bestI  = 0;

  for(i = 0; i < n; i++)
    {
      avgTime += times[i];
      avgLH   += likelihoods[i];
      if(likelihoods[i] > bestLH)
	{
	  bestLH = likelihoods[i];
	  bestI  = i;
	}
    }

  printf("Average Time per dataset: %f\n", (avgTime / n));
  printf("Average Likelihood: %f\n", (avgLH / n));
  printf("\n");
  printf("Best Likelihood in run number %d: likelihood %f\n", bestI, bestLH);

  infoFile = fopen(infoFileName, "a");
  fprintf(infoFile, "Average Time per dataset: %f\n", (avgTime / n));
  fprintf(infoFile, "Average Likelihood: %f\n", (avgLH / n));
  fprintf(infoFile, "\n");
  fprintf(infoFile, "Best Likelihood in run number %d: likelihood %f\n", bestI, bestLH);
  fclose(infoFile);

  if(adef->checkpoints)   
     printf("Checkpoints written to:                 %sRAxML_checkpoint.%s.RUN.%d to %d\n", workdir, run_id, 0, n-1);  
   
  printf("Parsimony starting trees written to:    %s.RUN.%d to %d\n", permFileName, 0, n-1);   	            
  printf("Final trees written to:                 %s.RUN.%d to %d\n", resultFileName,  0, n-1);  
  printf("Execution Log Files written to:         %sRAxML_log.%s.RUN.%d to %d\n", workdir, run_id, 0, n-1);   
  printf("Execution information file written to:  %s\n", infoFileName);


  free(times);
  free(likelihoods); 
}
